import abc
import functools
import logging
import os
import json
from dataclasses import dataclass
from typing import Dict, Iterable, Callable


import torch
import torch.utils.data
from diffusers.utils import is_accelerate_available
from diffusers.configuration_utils import ConfigMixin, register_to_config

if is_accelerate_available():
    import accelerate
else:
    accelerate = None

from .. import constant
from ..conversion import get_obj_from_str
from ..logger.logging import init_logger
from ..random._seeds import seed_every_thing
from ..decorators import Register
from ..auto import (
    # ConfigMixin,
    auto_cls_from_pretrained,
    config_dataclass_wrapper,
    get_version,
)

logger = logging.getLogger(__name__)


_trainer_register = Register('trainer')

register_trainer = _trainer_register.register()


def list_trainers():
    return list(_trainer_register.keys())


def get_trainer_builder(name: str):
    return _trainer_register[name]


@config_dataclass_wrapper(config_name='train_config.json')
@dataclass
class TrainingConfig:
    batch_size: int
    test_batch_size: int = None
    max_epochs: int = None
    max_steps: int = None
    checkpointing_epochs: int = None
    checkpointing_steps: int = None
    validation_epochs: int = None
    validation_steps: int = None
    saving_epochs: int = None
    saving_steps: int = None
    watching_epochs: int = None
    watching_steps: int = None

    def __post_init__(self) -> None:
        if self.test_batch_size is None:
            logger.warning(
                f"`test_batch_size` is None, auto set to `batch_size` ({self.batch_size})."
            )
            self.test_batch_size = self.batch_size
        if (self.max_epochs is None) and (self.max_steps is None):
            logger.warning(
                f"Both `max_epochs` and `max_steps` are None. "
                f"`max_epochs` is automatically set to 10000."
            )
            self.max_epochs = 10000
        elif (self.max_epochs is not None) and (self.max_steps is not None):
            logger.warning(
                f"Both `max_epochs` and `max_steps` are given. "
                f"Training will end when either limit is reached."
            )
        if (self.checkpointing_epochs is None) and (self.checkpointing_steps is None):
            logger.warning(
                f"Both `checkpointing_epochs` and `checkpointing_steps` are None. "
                f"No checkpoints will be saved during the training."
            )
        elif (self.checkpointing_epochs is not None) and (
            self.checkpointing_steps is not None
        ):
            logger.warning(
                f"Both `checkpointing_epochs` and `checkpointing_steps` are given. "
                f"All checkpoints meeting the criteria will be saved."
            )
        if (self.validation_epochs is None) and (self.validation_steps is None):
            logger.warning(
                f"Both `validation_epochs` and `validation_steps` are None. "
                f"No validation will be performed during the training."
            )
        elif (self.validation_epochs is not None) and (
            self.validation_steps is not None
        ):
            logger.warning(
                f"Both `validation_epochs` and `validation_steps` are given. "
                f"All validation meeting the criteria will be performed."
            )
        if (self.saving_epochs is None) and (self.saving_steps is None):
            logger.warning(
                f"Both `saving_epochs` and `saving_steps` are None. "
                f"No states will be saved during the training."
            )
        elif (self.saving_epochs is not None) and (self.saving_steps) is not None:
            logger.warning(
                f"Both `saving_epochs` and `saving_steps` are given. "
                f"All states meeting the criteria will be saved."
            )
        if (self.watching_epochs is None) and (self.watching_steps is None):
            logger.warning(
                f"Both `saving_epochs` and `saving_steps` are None. "
                f"No variables will be saved during the training."
            )
        elif (self.watching_epochs is not None) and (self.watching_steps is not None):
            logger.warning(
                f"Both `saving_epochs` and `saving_steps` are given. "
                f"all variables meeting the criteria will be saved."
            )


@config_dataclass_wrapper(config_name='train_outputfiles.json')
@dataclass
class TrainingOutputFilesManager:
    logging_file: str | os.PathLike = "debug.log"
    version_file: str | os.PathLike = "version.json"
    logging_dir: str | os.PathLike = "logs"
    tensorboard_dir: str | os.PathLike = "tensorboard"
    wandb_dir: str | os.PathLike = "wandb"
    checkpoints_dir: str | os.PathLike = "checkpoints"


@config_dataclass_wrapper(config_name='train_flag.json')
@dataclass
class TrainingFlag:
    step: int = 0
    epoch: int = 0


@dataclass
class TrainingDatasetManager:
    training_dataset: torch.utils.data.Dataset = None
    validation_dataset: torch.utils.data.Dataset = None
    training_dataloader: torch.utils.data.DataLoader = None
    validation_dataloader: torch.utils.data.DataLoader = None


class TrainerMixin(abc.ABC, ConfigMixin):

    config_name = "config.json"
    output_files_manager = TrainingOutputFilesManager()

    def __init__(
        self,
        output_dir: str | os.PathLike,
        project: str,
        name: str,
        seed: int = 0,
    ) -> None:
        super().__init__()
        self._set_seed(seed)
        self.project = project
        self.name = name
        self.output_dir = os.path.join(output_dir, project, name)
        self.writer_dir = os.path.join(output_dir, project)
        self.flag = TrainingFlag()
        self.store_to_log = dict()

    def _set_seed(self, seed: int) -> None:
        self.seed = seed
        seed_every_thing(seed)

    def __post_init__(self) -> None:
        if self.is_local_main_process:
            self.save_config(self.output_dir)
            self.save_version()
            config_dict = self._internal_dict if hasattr(self, "_internal_dict") else {}
            if self.accelerator is not None:
                self.accelerator.init_trackers(
                    project_name=self.project,
                    init_kwargs={
                        "wandb": {
                            "name": self.name,
                            "dir": self.writer_dir,
                            "config": config_dict,
                        }
                    },
                )
        logger.debug("Save config and version.")

    def save_version(self) -> dict:
        version = get_version()
        with open(
            os.path.join(self.output_dir, self.output_files_manager.version_file), "w"
        ) as file:
            json.dump(version, file, indent=2)

    def register_logger(self, logger_config: Dict = {}) -> None:
        os.makedirs(
            os.path.join(self.output_dir, self.output_files_manager.logging_dir),
            exist_ok=True,
        )
        if self.accelerator is not None:
            filename = f"rank{self.accelerator.state.process_index}.{self.output_files_manager.logging_file}"
        else:
            filename = self.output_files_manager.logging_file
        if not self.is_local_main_process:
            logger_config["console_level"] = "CRITICAL"
        init_logger(
            os.path.join(
                self.output_dir, self.output_files_manager.logging_dir, filename
            ),
            **logger_config,
        )

    def register_accelerator(self, accelerator_config: Dict = None) -> None:
        if accelerate is None or accelerator_config is None:
            self.accelerator = None
            self.use_ddp = False
        else:
            self.accelerator = accelerate.Accelerator(**accelerator_config)
            self.use_ddp = True

    def register_training_config(self, training_config: Dict) -> None:
        self.training_config = TrainingConfig(**training_config)

    def register_training_dataset_manager(
        self,
        training_dataset: torch.utils.data.Dataset = None,
        validation_dataset: torch.utils.data.Dataset = None,
        training_dataloader: torch.utils.data.DataLoader = None,
        validation_dataloader: torch.utils.data.DataLoader = None,
    ) -> None:
        self.dataset_manager = TrainingDatasetManager(
            training_dataset,
            validation_dataset,
            training_dataloader,
            validation_dataloader,
        )

    def register_optimizer(self, params: Iterable, optimizer_config: Dict) -> None:
        optimizer_name = optimizer_config.pop("name", "torch.optim.Adam")
        optimizer_cls = get_obj_from_str(optimizer_name)
        self.optimizer = optimizer_cls(params, **optimizer_config)
        logger.debug(f"Load {optimizer_name}.")

    def register_lr_scheduler(self, lr_scheduler_config: Dict) -> None:
        lr_scheduler_name = lr_scheduler_config.pop(
            "name", "torch.optim.lr_scheduler.LambdaLR"
        )
        set_to_max_epoch = lr_scheduler_config.pop("set_to_max_epoch", None)
        set_to_max_iter = lr_scheduler_config.pop("set_to_max_iter", None)
        if set_to_max_epoch is not None:
            lr_scheduler_config[set_to_max_epoch] = self.training_config.max_epochs
        if set_to_max_iter is not None:
            lr_scheduler_config[set_to_max_iter] = self.training_config.max_steps
        lr_scheduler_cls = get_obj_from_str(lr_scheduler_name)
        self.lr_scheduler = lr_scheduler_cls(self.optimizer, **lr_scheduler_config)
        logger.debug(f"Load {lr_scheduler_name}.")

    @property
    def is_local_main_process(self) -> bool:
        if not self.accelerator:
            return True
        return self.accelerator.state.process_index == 0

    def wait_for_everyone(self):
        if self.accelerator is not None:
            self.accelerator.wait_for_everyone()

    def no_ending(self):
        if self.training_config.max_epochs is not None:
            if self.flag.epoch < self.training_config.max_epochs:
                return True
        if self.training_config.max_steps is not None:
            if self.flag.step < self.training_config.max_steps:
                return True
        return False

    @abc.abstractmethod
    def auto_set_to_train_mode(self):
        raise NotImplementedError(
            "Please implement the `auto_set_to_train_mode` method."
        )

    @abc.abstractmethod
    def auto_set_to_eval_mode(self):
        raise NotImplementedError(
            "Please implement the `auto_set_to_eval_mode` method."
        )

    @abc.abstractmethod
    def auto_save_checkpoint(self):
        raise NotImplementedError("Please implement the `auto_save_checkpoint` method.")

    @abc.abstractmethod
    def auto_save_training_status(self):
        raise NotImplementedError(
            "Please implement the `auto_save_training_status` method."
        )

    @abc.abstractmethod
    def auto_validation(self):
        raise NotImplementedError("Please implement the `auto_validation` method.")

    @abc.abstractmethod
    def auto_watching(self):
        raise NotImplementedError("Please implement the `auto_watching` method.")

    @abc.abstractmethod
    def auto_load_status(self, checkpoint_path):
        raise NotImplementedError("Please implement the `auto_load_status` method.")

    @torch.no_grad()
    def _save_checkpoint(self):
        self.wait_for_everyone()
        if self.is_local_main_process:
            self.auto_set_to_eval_mode()
            self.auto_save_checkpoint()
            self.auto_set_to_train_mode()
        self.wait_for_everyone()

    @torch.no_grad()
    def _save_training_status(self):
        self.wait_for_everyone()
        if self.is_local_main_process:
            self.auto_set_to_eval_mode()
            self.auto_save_training_status()
            self.auto_set_to_train_mode()
        self.wait_for_everyone()

    @torch.no_grad()
    def _validation(self):
        self.wait_for_everyone()
        self.auto_set_to_eval_mode()
        self.auto_validation()
        self.auto_set_to_train_mode()
        self.wait_for_everyone()

    @torch.no_grad()
    def _watching(self):
        self.wait_for_everyone()
        if self.is_local_main_process:
            self.auto_set_to_eval_mode()
            self.auto_watching()
            self.auto_set_to_train_mode()
        self.wait_for_everyone()

    @classmethod
    def from_yaml_config(cls, config_path: os.PathLike):
        import yaml

        with open(config_path, "r") as config_file:
            config = yaml.safe_load(config_file)
        return cls.from_config(config)

    @classmethod
    def from_json_config(cls, config_path: os.PathLike):
        import json

        with open(config_path, "r") as config_file:
            config = json.load(config_file)
        return cls.from_config(config)

    @classmethod
    def from_ini_config(cls, config_path: os.PathLike):
        import configparser

        config = configparser.ConfigParser()
        config.read(config_path)
        config = {section: dict(config.items(section)) for section in config.sections()}
        return cls.from_config(config)

    @classmethod
    def from_config_file(cls, config_path: os.PathLike):
        from pathlib import Path

        extension = Path(config_path).suffix.lower()
        if extension in constant.ConfigSuffix.json:
            return cls.from_json_config(config_path)
        if extension in constant.ConfigSuffix.yaml:
            return cls.from_yaml_config(config_path)
        if extension in constant.ConfigSuffix.ini:
            return cls.from_ini_config(config_path)
        raise NotImplementedError(
            f"Unknown suffix '{extension}' in path '{config_path}'."
        )

    @classmethod
    def from_output_dir(cls, dir_path: os.PathLike):
        config_path = os.path.join(dir_path, cls.config_name)
        self = cls.from_config_file(config_path)
        self.auto_load_status()

        logger.info(f"Warm up for epoch={self.flag.epoch} and step={self.flag.step}.")
        if self.flag.step == 0 and self.flag.epoch == 0:
            return self

        expect_epoch = self.flag.step // len(self.dataset_manager.training_dataloader)
        if expect_epoch != self.flag.epoch:
            raise RuntimeError("`current_epoch` and `current_step` mismatch.")

        return self


def register_to_run_one_epoch(only_training: bool = False):

    def one_epoch_func_decorator(one_epoch_func: Callable):

        @functools.wraps(one_epoch_func)
        def run_one_epoch(self: TrainerMixin, *args, **kwargs):
            logger.debug(f"Start Epoch {self.flag.epoch}.")
            result = one_epoch_func(self, *args, **kwargs)
            logger.debug(f"End Epoch {self.flag.epoch}.")

            self.flag.epoch += 1

            if only_training:
                return result

            if self.training_config.checkpointing_epochs is not None:
                if self.flag.epoch % self.training_config.checkpointing_epochs == 0:
                    self._save_checkpoint()

            if self.training_config.saving_epochs is not None:
                if self.flag.epoch % self.training_config.saving_epochs == 0:
                    self._save_training_status()

            if self.training_config.validation_epochs is not None:
                if self.flag.epoch % self.training_config.validation_epochs == 0:
                    self._validation()

            if self.training_config.watching_epochs is not None:
                if self.flag.epoch % self.training_config.watching_epochs == 0:
                    self._watching()

            return result

        return run_one_epoch

    return one_epoch_func_decorator


def register_to_run_one_iteration(only_training: bool = False):

    def one_iteration_func_decorator(one_iteration_func: Callable):

        @functools.wraps(one_iteration_func)
        def run_one_iteration(self: TrainerMixin, *args, **kwargs):
            result = one_iteration_func(self, *args, **kwargs)

            self.flag.step += 1
            if hasattr(self, "iter_process_bar") and self.iter_process_bar is not None:
                self.iter_process_bar.update(1)

            if only_training:
                return result

            if self.training_config.checkpointing_steps is not None:
                if self.flag.step % self.training_config.checkpointing_steps == 0:
                    self._save_checkpoint()

            if self.training_config.saving_steps is not None:
                if self.flag.step % self.training_config.saving_steps == 0:
                    self._save_training_status()

            if self.training_config.validation_steps is not None:
                if self.flag.step % self.training_config.validation_steps == 0:
                    self._validation()

            if self.training_config.watching_steps is not None:
                if self.flag.step % self.training_config.watching_steps == 0:
                    self._watching()

            return result

        return run_one_iteration

    return one_iteration_func_decorator


def auto_trainer_from_pretrained(path: str, **kwargs):

    return auto_cls_from_pretrained(_trainer_register, TrainerMixin, path, **kwargs)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Callable, Dict, List, Tuple, Optional

from torchvision.datasets import DatasetFolder
from torchvision.datasets.folder import default_loader, IMG_EXTENSIONS


class LabelDatasetFolder(DatasetFolder):
    """A label data loader.

    The subfolder of the root should be named as the label number.

    Args:
        root (string): Root directory path.
        loader (callable): A function to load a sample given its path.
        extensions (tuple[string]): A list of allowed extensions.
            both extensions and is_valid_file should not be passed.
        transform (callable, optional): A function/transform that takes in
            a sample and returns a transformed version.
            E.g, ``transforms.RandomCrop`` for images.
        target_transform (callable, optional): A function/transform that takes
            in the target and transforms it.
        is_valid_file (callable, optional): A function that takes path of a file
            and check if the file is a valid file (used to check of corrupt files)
            both extensions and is_valid_file should not be passed.

     Attributes:
        classes (list): List of the class names sorted alphabetically.
        class_to_idx (dict): Dict with items (class_name, class_index).
        samples (list): List of (sample path, class_index) tuples
        targets (list): The class_index value for each image in the dataset
    """

    def __init__(
        self,
        root: str,
        loader: Callable[[str], Any],
        extensions: Tuple[str, ...] | None = None,
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        is_valid_file: Callable[[str], bool] | None = None,
    ) -> None:
        super().__init__(
            root, loader, extensions, transform, target_transform, is_valid_file
        )

        classes, class_to_idx = self.find_classes(self.root)
        samples = self.make_dataset(self.root, class_to_idx, extensions, is_valid_file)

        self.loader = loader
        self.extensions = extensions

        self.classes = classes
        self.class_to_idx = class_to_idx
        self.samples = samples
        self.targets = [s[1] for s in samples]

    def find_classes(self, directory: str) -> Tuple[List[str], Dict[str, int]]:
        classes = sorted(
            (
                entry.name
                for entry in os.scandir(directory)
                if entry.is_dir() and entry.name.isalnum()
            ),
            key=lambda x: int(x),
        )
        if not classes:
            raise FileNotFoundError(f"Couldn't find any class folder in {directory}.")

        class_to_idx = {cls_name: int(cls_name) for i, cls_name in enumerate(classes)}
        return classes, class_to_idx


class LabelImageFolder(LabelDatasetFolder):
    """A generic data loader where the images are arranged in this way by default: ::

        root/0/xxx.png
        root/0/xxy.png

        root/1/123.png
        root/1/nsdf3.png

    This class inherits from :class:`LabelDatasetFolder` so
    the same methods can be overridden to customize the dataset.

    Args:
        root (string): Root directory path.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        loader (callable, optional): A function to load an image given its path.
        is_valid_file (callable, optional): A function that takes path of an Image file
            and check if the file is a valid file (used to check of corrupt files)

     Attributes:
        classes (list): List of the class names sorted alphabetically.
        class_to_idx (dict): Dict with items (class_name, class_index).
        imgs (list): List of (image path, class_index) tuples
    """

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        loader: Callable[[str], Any] = default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ):
        super().__init__(
            root,
            loader,
            IMG_EXTENSIONS if is_valid_file is None else None,
            transform=transform,
            target_transform=target_transform,
            is_valid_file=is_valid_file,
        )
        self.imgs = self.samples



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Tuple, Literal
from dataclasses import dataclass, field

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from .utils import FontManager, AxesManager, Painter
from ..auto.dataclass_utils import config_dataclass_wrapper


@config_dataclass_wrapper()
@dataclass
class LineInfo:
    x: list | np.ndarray | pd.Series
    y: list | np.ndarray | pd.Series
    label: str | None = None
    color: str | None = None
    marker: str = "o"
    alpha: float = 0.9
    linewidth: float = 1.0
    linestyle: str = "-"
    markersize: float = 7.0


class LineChartPainter(Painter):
    def __init__(self, fig_size: Tuple[int, int], font_manager: FontManager):
        super().__init__(fig_size=fig_size, font_manager=font_manager)

    def draw(self, line_info: LineInfo):
        self.ax.plot(
            line_info.x,
            line_info.y,
            label=line_info.label,
            color=line_info.color,
            marker=line_info.marker,
            alpha=line_info.alpha,
            linewidth=line_info.linewidth,
            linestyle=line_info.linestyle,
            markersize=line_info.markersize,
        )



import os
import json
import logging
from typing import Optional

from .path import get_safe_save_path
from ..format import format_as_yaml

logger = logging.getLogger(__name__)


def safe_save_as_yaml(obj, save_path: str):
    """Save the data in yaml format.

    Args:
        obj (Any): The objective to save.
        save_dir (str): The directory path.
        save_name (Optional[str], optional): The file name for the data to save. Defaults to None.
    """

    s = format_as_yaml(obj)
    save_path = get_safe_save_path(save_path)

    with open(save_path, 'w') as f:
        f.write(s)

    logger.info(f'Save yaml to {save_path}.')


def safe_save_torchobj(obj, save_path: str):
    """Save the obj by using torch.save function.

    Args:
        obj (Any): The objective to save.
        save_dir (str): The directory path.
        save_name (Optional[str], optional): The file name for the objective to save. Defaults to None.
    """

    save_path = get_safe_save_path(save_path)

    import torch

    torch.save(obj, save_path)

    logger.info(f'Save torch object to {save_path}.')


def safe_save_as_pickle(obj, save_path: str):
    """Save the data in pickle format.

    Args:
        obj (Any): The objective to save.
        save_dir (str): The directory path.
        save_name (Optional[str], optional): The file name for the data to save. Defaults to None.
    """

    save_path = get_safe_save_path(save_path)

    import pickle

    with open(save_path, 'wb') as f:
        pickle.dump(obj, f)


def safe_save_csv(obj, save_path: str):
    """Save the data in csv format.

    Args:
        obj (Any): The objective to save.
        save_dir (str): The directory path.
        save_name (Optional[str], optional): The file name for the data to save. Defaults to None.
    """
    import pandas as pd

    if not isinstance(obj, pd.DataFrame):
        df = pd.DataFrame(obj)

    save_path = get_safe_save_path(save_path)
    df.to_csv(save_path, index=None)

    logger.info(f'Save csv to {save_path}.')


def safe_save_as_json(obj, save_path: str):
    """Save the data in json format.

    Args:
        obj (Any): The objective to save.
        save_dir (str): The directory path.
        save_name (Optional[str], optional): The file name for the data to save. Defaults to None.
    """

    save_path = get_safe_save_path(save_path)
    with open(save_path, 'w') as f:
        json.dump(obj, f, indent=2, sort_keys=True)

    logger.info(f'Save json to {save_path}.')



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import functools
from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional, Union, List, Dict, Any, Literal

import torch


from diffusers.utils import is_accelerate_available

if is_accelerate_available():
    from accelerate import Accelerator

from ..auto import PipelineMixin, config_dataclass_wrapper, only_main_process
from ..random import seed_every_thing
from .optimization import OptimizationManager, SingleOptimizationManager

logger = logging.getLogger(__name__)


def register_evaluate_function(func):

    @functools.wraps(func)
    def wrapper(self: TrainPipelineMixin, *args, **kwargs):
        if self.training is not False:
            self.set_to_eval_mode()
        return func(self, *args, **kwargs)

    return wrapper


def register_train_function(func):

    @functools.wraps(func)
    def wrapper(self: TrainPipelineMixin, *args, **kwargs):
        if self.training is not True:
            self.set_to_train_mode()
        # logger.critical(f'execute {func.__name__} {self.execute_counts}')
        return func(self, *args, **kwargs)

    return wrapper


class TrainPipelineMixin(PipelineMixin):

    optimization_manager: OptimizationManager = OptimizationManager()
    _optimization_save_folder = 'optimization'

    _training = None

    def __init__(
        self,
        output_dir: str,
        seed: int = 0,
        accelerator: Optional["Accelerator"] = None,
    ) -> None:
        super().__init__()
        self.register_accelerator(accelerator)
        self.register_save_values(execute_counts=self.execute_counts)
        self.output_dir = output_dir
        self._set_seed(seed)

        self._logger = logging.getLogger(self.__class__.__name__)

    @property
    def logger(self):
        return self._logger

    @property
    def training(self):
        return self._training

    @staticmethod
    def register_training_function(func):
        return register_train_function(func)

    @staticmethod
    def register_evaluate_function(func):
        return register_evaluate_function(func)

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, **kwargs):
        output_dir = kwargs.pop('output_dir', pretrained_model_name_or_path)
        self = super().from_pretrained(
            pretrained_model_name_or_path, output_dir=output_dir, **kwargs
        )
        optimization_save_path = os.path.join(
            self.output_dir, self._optimization_save_folder
        )
        self.optimization_manager.load_state_dict_from_file(optimization_save_path)

        return self

    @only_main_process
    def save_pretrained(self, save_directory: str | os.PathLike):
        optimization_save_path = os.path.join(
            self.output_dir, self._optimization_save_folder
        )
        self.optimization_manager.save_state_dict_to_file(optimization_save_path)
        super().save_pretrained(save_directory)

    def _set_seed(self, seed):
        self.seed = seed
        seed_every_thing(seed)

    def register_optimization(self, tag, params, optimization_config):
        self.optimization_manager.register_optimization(
            tag, params, optimization_config
        )

    # def train(self, training: bool = True):
    #     self._training = training
    #     if training:
    #         self.set_to_train_mode()
    #     else:
    #         self.set_to_eval_mode()

    # def eval(self):
    #     self.train(False)

    @abstractmethod
    def set_to_train_mode(self):
        pass

    @abstractmethod
    def set_to_eval_mode(self):
        pass

    def accelerator_prepare(self, *args, device_placement=None):
        if self.accelerator is None:
            return args

        result = []
        for to_prepare in args:
            if isinstance(to_prepare, (OptimizationManager, SingleOptimizationManager)):
                to_prepare.accelerator_prepare(
                    self.accelerator, device_placement=device_placement
                )
            else:
                to_prepare = self.accelerator.prepare(to_prepare)
            result.append(to_prepare)

        return result

    def backward(self, loss):
        if self.accelerator is not None:
            self.accelerator.backward(loss)
        else:
            loss.backward()



# Copyright 2024 Sijin Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import requests
import time

logger = logging.getLogger(__name__)


def download_file(
    url: str,
    local_path: os.PathLike,
    stream: bool = True,
    mkdirs: bool = True,
    chunk_size: int = 8192,
    retry: int = 10,
    wait: float = 2.0,
):
    """Downloads a file from a specified URL and saves it to a local path.

    Args:
        url (str): The URL from which to download the file.
        local_path (os.PathLike): The local file system path where the file should be saved.
        stream (bool, optional): Whether to stream the download. Default is True, recommended for large files.
            Defaults to True.
        mkdirs (bool, optional): If True, create the directory for `local_path` if it doesn't exist.
            Defaults to True.
        chunk_size (int, optional): Size of chunks to download at a time in bytes. Defaults to 8192.
        retry (int, optional): Number of times to retry the download in case of failure. Defaults to 10.
        wait (float, optional): Time to wait between retries in seconds. Defaults to 2.0.

    Returns:
        bool: True if the file was successfully downloaded and saved, False otherwise.
    """
    dirname = os.path.dirname(local_path)
    if mkdirs and dirname != "":
        os.makedirs(dirname, exist_ok=True)
    for i in range(retry):
        try:
            with requests.get(url, stream=stream) as response:
                with open(local_path, "wb") as file:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        file.write(chunk)
            logger.info(f"Download {local_path} from {url}.")
            return True
        except Exception as e:
            logger.warning(
                f"Failed to download from {url}: {str(e)}, retry {i + 1} / {retry} after {wait} sec."
            )
            time.sleep(wait)
    logger.error(f"Failed to download from {url}.")
    if os.path.exists(local_path):
        os.remove(local_path)



import torch


def reparameterize(mu, std):

    eps = torch.randn_like(std)

    return eps * std + mu


def copy_or_set_(dest, source):
    """
    A workaround to respect strides of :code:`dest` when copying :code:`source`
    (https://github.com/geoopt/geoopt/issues/70)
    Parameters
    ----------
    dest : torch.Tensor
        Destination tensor where to store new data
    source : torch.Tensor
        Source data to put in the new tensor
    Returns
    -------
    dest
        torch.Tensor, modified inplace
    """
    if dest.stride() != source.stride():
        return dest.copy_(source)
    else:
        return dest.set_(source)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import yaml
import json
import logging
from typing import Optional
from collections import OrderedDict

logger = logging.getLogger(__name__)


def format_as_split_line(content: Optional[str] = None, length: int = 60) -> str:
    if content is None:
        return '-' * length
    if len(content) > length - 4:
        logger.info('The length of content is larger than the specified length.')
        length = len(content) + 4

    total_num = length - len(content) - 2
    left_num = total_num // 2
    right_num = total_num - left_num
    return '-' * left_num + f' {content} ' + '_' * right_num


def format_number(num: int, base: int = 1000):
    prefix = ''
    if num < 0:
        prefix = '-'
        num = -num

    for suffix in ['', 'k', 'M', 'G']:
        if num < base:
            return f'{num:.4f} {suffix}'
        num /= base

    return f'{prefix}{num:.4f} T'


def format_time_invterval(t: float):
    if t <= 60:
        return f'{t: .4f} seconds'
    t /= 60
    if t <= 60:
        return f'{t: .4f} minutes'
    t /= 60
    if t <= 24:
        return f'{t: .4f} hours'
    t /= 24
    return f'{t: .4f} days'


def format_time(time_time: Optional[float] = None, pattern: str = '%Y%m%d-%H%M%S'):
    if time_time is None:
        time_time = time.time()
    return time.strftime(pattern, time.localtime(time_time))


yaml.add_representer(
    OrderedDict,
    lambda dumper, data: dumper.represent_mapping(
        'tag:yaml.org,2002:map', data.items()
    ),
)
yaml.add_representer(
    tuple, lambda dumper, data: dumper.represent_sequence('tag:yaml.org,2002:seq', data)
)


def format_as_yaml(obj) -> str:
    return yaml.dump(obj)


def format_as_json(obj, indent=2, sort_keys=True) -> str:
    return json.dumps(obj, indent=indent, sort_keys=sort_keys)



import os
import hashlib

package_name = 'phoenixcat'
package_name_upper = package_name.upper()

USER_NAME = os.getlogin()
USER_NAME_HASH = hashlib.md5(USER_NAME.encode()).hexdigest()

default_home = os.path.join(os.path.expanduser("~"), ".cache", package_name)
HOME = os.path.expanduser(
    os.getenv(
        f"{package_name_upper}_HOME",
        default_home,
    )
)


default_assets_cache_path = os.path.join(HOME, "assets")

ASSETS_CACHE = os.getenv(
    f"{package_name_upper}_ASSETS_CACHE", default_assets_cache_path
)

os.makedirs(ASSETS_CACHE, exist_ok=True)


default_config_path = os.path.join(HOME, "config")

CONFIG_HOME = os.getenv(f"{package_name_upper}_CONFIG_PATH", default_config_path)

os.makedirs(CONFIG_HOME, exist_ok=True)


class ConfigSuffix:
    json = ["json", "JSON", ".json", ".JSON"]
    yaml = ["yaml", "YAML", "yml", "YML", ".yaml", ".YAML", ".yml", ".YML"]
    ini = ["ini", "INI", ".ini", ".INI"]
    py = ["py", "PY", ".py", ".PY"]



import logging
from typing import Dict, List, Any, Literal

from diffusers.configuration_utils import register_to_config

from ..trainer_utils import TrainerMixin
from ...decorators import Register
from ...configuration import ConfigMixin, extract_init_dict, auto_cls_from_pretrained


logger = logging.getLogger(__name__)

_callback_register = Register('callback')

register_callback = _callback_register.register()


def list_callbacks():
    return list(_callback_register.keys())


def get_callback_builder(name: str):
    return _callback_register[name]


class CallbackMixin(ConfigMixin):

    config_name = 'callback.json'

    def __init__(self):
        pass

    # def __enter__(self):
    #     pass

    # def __exit__(self, *err):
    #     """Release resources here if have any."""

    def begin(self, trainer: TrainerMixin):
        pass

    def step_begin(self, trainer: TrainerMixin):
        pass

    def step_end(self, trainer: TrainerMixin):
        pass

    def epoch_begin(self, trainer: TrainerMixin):
        pass

    def epoch_end(self, trainer: TrainerMixin):
        pass

    def end(self, trainer: TrainerMixin):
        pass

    def on_train_begin(self, trainer: TrainerMixin):
        """
        Called once before the network training.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.begin(trainer)

    def on_train_epoch_begin(self, trainer: TrainerMixin):
        """
        Called before each training epoch begin.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.epoch_begin(trainer)

    def on_train_epoch_end(self, trainer: TrainerMixin):
        """
        Called after each training epoch end.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.epoch_end(trainer)

    def on_train_step_begin(self, trainer: TrainerMixin):
        """
        Called before each training step begin.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.step_begin(trainer)

    def on_train_step_end(self, trainer: TrainerMixin):
        """
        Called after each training step end.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.step_end(trainer)

    def on_train_end(self, trainer: TrainerMixin):
        """
        Called after training end.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.end(trainer)

    def on_eval_begin(self, trainer: TrainerMixin):
        """
        Called before eval begin.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.begin(trainer)

    def on_eval_epoch_begin(self, trainer: TrainerMixin):
        """
        Called before eval epoch begin.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.epoch_begin(trainer)

    def on_eval_epoch_end(self, trainer: TrainerMixin):
        """
        Called after eval epoch end.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.epoch_end(trainer)

    def on_eval_step_begin(self, trainer: TrainerMixin):
        """
        Called before each eval step begin.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.step_begin(trainer)

    def on_eval_step_end(self, trainer: TrainerMixin):
        """
        Called after each eval step end.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.step_end(trainer)

    def on_eval_end(self, trainer: TrainerMixin):
        """
        Called after eval end.

        Args:
            trainer (TrainerMixin): Include some information of the model.
        """
        self.end(trainer)


@register_callback
class ComposeCallback(CallbackMixin):
    """
    Sequential execution of callback functions.

    Execute Callback functions at certain points.
    """

    @register_to_config
    def __init__(self, callbacks_init_list: List[Dict[Literal['name', 'kwargs'], Any]]):
        self._callbacks = []
        for cb_args in callbacks_init_list:
            if 'name' not in cb_args and 'kwargs' not in cb_args:
                raise RuntimeError(f'Parameters of `callbacks_init_list` error!')
            factory = get_callback_builder(cb_args['name'])
            cb = factory(**cb_args['kwargs'])
            self._callbacks.append(cb)

    @classmethod
    def create_from_callbacks(cls, callbacks: CallbackMixin, **kwargs):

        _callbacks = []
        if isinstance(callbacks, CallbackMixin):
            _callbacks.append(callbacks)
        elif isinstance(callbacks, list):
            for cb in callbacks:
                if not isinstance(cb, CallbackMixin):
                    raise TypeError(
                        "When the 'callbacks' is a list, the elements in "
                        "'callbacks' must be Callback functions."
                    )
                _callbacks.append(cb)
        elif callbacks is not None:
            raise TypeError("The 'callbacks' is not a Callback or a list of Callback.")

        cb_list = []
        for inner_cb in _callbacks:
            inner_cb_name = inner_cb.__class__.__name__
            inner_cb_kwargs = extract_init_dict(inner_cb)[0]
            cb_list.append({'name': inner_cb_name, 'kwargs': inner_cb_kwargs})

        return cls(cb_list)

    def begin(self, trainer):
        """Called once before network train or eval."""
        for cb in self._callbacks:
            cb.begin(trainer)

    def epoch_begin(self, trainer):
        """Called before each epoch begin."""
        for cb in self._callbacks:
            cb.epoch_begin(trainer)

    def epoch_end(self, trainer):
        """Called after each epoch finished."""
        for cb in self._callbacks:
            cb.epoch_end(trainer)

    def step_begin(self, trainer):
        """Called before each step begin."""
        for cb in self._callbacks:
            cb.step_begin(trainer)

    def step_end(self, trainer):
        """Called after each step finished."""
        for cb in self._callbacks:
            cb.step_end(trainer)

    def end(self, trainer):
        """Called once after network train or eval."""
        for cb in self._callbacks:
            cb.end(trainer)

    def on_train_begin(self, trainer):
        """Called before network train."""
        for cb in self._callbacks:
            cb.on_train_begin(trainer)

    def on_train_epoch_begin(self, trainer):
        """Called before each train epoch begin."""
        for cb in self._callbacks:
            cb.on_train_epoch_begin(trainer)

    def on_train_epoch_end(self, trainer):
        """Called after each train epoch finished."""
        for cb in self._callbacks:
            cb.on_train_epoch_end(trainer)

    def on_train_step_begin(self, trainer):
        """Called before each train step begin."""
        for cb in self._callbacks:
            cb.on_train_step_begin(trainer)

    def on_train_step_end(self, trainer):
        """Called after each train step finished."""
        for cb in self._callbacks:
            cb.on_train_step_end(trainer)

    def on_train_end(self, trainer):
        """Called after network train end."""
        for cb in self._callbacks:
            cb.on_train_end(trainer)

    def on_eval_begin(self, trainer):
        """Called before network eval."""
        for cb in self._callbacks:
            cb.on_eval_begin(trainer)

    def on_eval_epoch_begin(self, trainer):
        """Called before eval epoch begin."""
        for cb in self._callbacks:
            cb.on_eval_epoch_begin(trainer)

    def on_eval_epoch_end(self, trainer):
        """Called after eval epoch finished."""
        for cb in self._callbacks:
            cb.on_eval_epoch_end(trainer)

    def on_eval_step_begin(self, trainer):
        """Called before each eval step begin."""
        for cb in self._callbacks:
            cb.on_eval_step_begin(trainer)

    def on_eval_step_end(self, trainer):
        """Called after each eval step finished."""
        for cb in self._callbacks:
            cb.on_eval_step_end(trainer)

    def on_eval_end(self, trainer):
        """Called after network eval end."""
        for cb in self._callbacks:
            cb.on_eval_end(trainer)


def auto_trainer_from_pretrained(path: str, **kwargs):

    return auto_cls_from_pretrained(_callback_register, CallbackMixin, path, **kwargs)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from diffusers.configuration_utils import register_to_config

from .classifier_utils import ExternalClassifier
from ..modeling_utils import register_model


@register_model
class TorchvisionClassifier(ExternalClassifier):

    ignore_for_config = ['weights']

    @register_to_config
    def __init__(
        self,
        arch_name: str,
        num_classes: int,
        resolution=224,
        weights=None,
        arch_kwargs={},
    ) -> None:

        if weights is not None:
            arch_kwargs['weights'] = weights

        super().__init__(
            arch_name=arch_name,
            pkg_name='torchvision.models',
            num_classes=num_classes,
            resolution=resolution,
            arch_kwargs=arch_kwargs,
        )


@register_model
class ResNeStClassifier(ExternalClassifier):

    ignore_for_config = ['weights']

    @register_to_config
    def __init__(
        self,
        arch_name: str,
        num_classes: int,
        resolution=224,
        weights=None,
        arch_kwargs={},
    ) -> None:

        if weights is not None:
            arch_kwargs['weights'] = weights

        super().__init__(
            arch_name=arch_name,
            pkg_name='resnest.torch',
            num_classes=num_classes,
            resolution=resolution,
            arch_kwargs=arch_kwargs,
        )



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="phoenixcat",
    version="0.6.0",
    author="Hongyao Yu, Sijin Yu",
    author_email="chrisqcwx@gmail.com, eeyusijin@mail.scut.edu.cn",
    description="An enhanced infrastructure library for PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Chrisqcwx/phoenixcat",
    package_dir={"": "src"},
    packages=find_packages("src"),
    license="Apache 2.0 License",
    python_requires=">=3.8",
    install_requires=["torch", "torchvision", "diffusers"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Utilities",
    ],
    entry_points={"console_scripts": ["phoenixcat=phoenixcat.cli:main"]},
)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import random

import numpy as np

logger = logging.getLogger(__name__)

_ALL_LOGITS = '0123456789qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM'


def random_string(length: int = 6):
    """Generate a random string with the specified length.

    Args:
        length (int, optional): The string length. Defaults to 6.

    Returns:
        str: The randomly generated string.
    """
    results = random.choices(_ALL_LOGITS, k=length)
    results = ''.join(results)
    logger.info(f'Generate random string `{results}`.')
    return results



from .format import (
    format_as_split_line,
    format_as_json,
    format_as_yaml,
    format_number,
    format_time,
    format_time_invterval,
)
from .latex import (
    DataPoint,
    ListDataPoint,
    TableDataPoint,
    ListMultiCellInfo,
    ListMultiRowInfo,
    TableRow,
    register_highlight_format,
    list_all_highlight_format,
)
from .font_color import get_colored_str



import functools
from typing import Mapping
from collections import OrderedDict
from diffusers.utils.outputs import BaseOutput as HF_BaseOutput
from dataclasses import dataclass, is_dataclass


def parallel_enable_dataclass(cls):
    """
    A decorator that extends dataclass to accept either keyword arguments or a dictionary.
    """
    # cls = dataclass(cls)

    if not is_dataclass(cls):
        cls = dataclass(cls)

    # Modify the __init__ method to accept a dictionary as an argument
    original_init = cls.__init__

    @functools.wraps(original_init)
    def new_init(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], Mapping):
            init_dict = args[0]
            original_init(self, **init_dict)
        else:
            original_init(self, *args, **kwargs)

    cls.__init__ = new_init
    return cls


# @parallel_enable_dataclass
# class BaseOutput(HF_BaseOutput):
#     aa: int
#     bb: str


# BaseOutput()



from .config_utils import ConfigMixin, extract_init_dict, auto_cls_from_pretrained

from .dataclass_utils import config_dataclass_wrapper, dict2dataclass

from .pipeline_utils import PipelineMixin, register_to_pipeline_init
from .version import get_current_commit_hash, get_version

from .autosave_utils import (
    is_json_serializable,
    AutoSaver,
    register_from_pretrained,
    register_save_pretrained,
    get_init_parameters,
    split_init_other_parameters,
    auto_register_save_load,
    auto_create_cls,
)

from .accelerater_utils import (
    AccelerateMixin,
    only_local_main_process,
    only_main_process,
)

from .dumper import Dumper, DictDumper, ListDumper



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from dataclasses import dataclass
from typing import Callable, Optional

import torch
from torch import nn
from torchvision.transforms.functional import resize
from diffusers.utils.outputs import BaseOutput

from ..modeling_utils import ModelMixin, get_model_builder
from ..output import parallel_enable_dataclass
from ...auto import extract_init_dict


class BaseImageModel(ModelMixin):

    def __init__(self, resolution: int) -> None:
        super().__init__()

        self._resolution = resolution

    @property
    def resolution(self):
        return self._resolution

    def preprocess_images_input(self, images: torch.Tensor):

        if images.shape[-1] != self.resolution or images.shape[-2] != self.resolution:
            images = resize(images, (self.resolution, self.resolution), antialias=True)

    def __call__(self, images, *args, **kwargs):
        self.preprocess_images_input(images)
        return super(ModelMixin, self).__call__(images, *args, **kwargs)


@parallel_enable_dataclass
@dataclass
class BaseImageClassifierOutput(BaseOutput):
    prediction: torch.Tensor
    feature: torch.Tensor


class BaseImageClassifier(BaseImageModel):

    WEIGHTS_NAME = 'classifier.bin'
    SAFETENSORS_WEIGHTS_NAME = 'classifier.safetensors'

    def __init__(
        self,
        resolution,
        num_classes,
        feature_dim,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(resolution, *args, **kwargs)
        self._num_classes = num_classes
        self._feature_dim = feature_dim

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def feature_dim(self):
        return self._feature_dim


def _reset_fc_impl(
    module: nn.Module, reset_num_classes: int = None, visit_fc_fn: Callable = None
):
    """Reset the output class num of nn.Linear and return the input feature_dim of nn.Linear.

    Args:
        module (nn.Module): The specific model structure.
        reset_num_classes (int, optional): The new output class num. Defaults to None.
        visit_fc_fn (Callable, optional): Other operations to the nn.Linear of the input module. Defaults to None.

    Returns:
        feature_dim (int): The input feature_dim of nn.Linear.
    """

    if isinstance(module, nn.Sequential):

        if len(module) == 0:
            raise RuntimeError('fail to implement')

        if isinstance(module[-1], nn.Linear):
            feature_dim = module[-1].weight.shape[-1]

            if (
                reset_num_classes is not None
                and reset_num_classes != module[-1].weight.shape[0]
            ):
                module[-1] = nn.Linear(feature_dim, reset_num_classes)

            if visit_fc_fn is not None:
                visit_fc_fn(module[-1])

            return feature_dim
        else:
            return _reset_fc_impl(module[-1], reset_num_classes)

    children = list(module.named_children())
    if len(children) == 0:
        raise RuntimeError('Fail to implement')
    attr_name, child_module = children[-1]
    if isinstance(child_module, nn.Linear):
        feature_dim = child_module.weight.shape[-1]

        if (
            reset_num_classes is not None
            and reset_num_classes != child_module.weight.shape[0]
        ):
            setattr(module, attr_name, nn.Linear(feature_dim, reset_num_classes))

        if visit_fc_fn is not None:
            visit_fc_fn(getattr(module, attr_name))

        return feature_dim
    else:
        return _reset_fc_impl(child_module, reset_num_classes)


def reset_fc_layer(
    module: nn.Module, reset_num_classes: int = None, visit_fc_fn: Callable = None
) -> int:
    return _reset_fc_impl(module, reset_num_classes, visit_fc_fn)


class ExternalClassifier(BaseImageClassifier):

    def __init__(
        self,
        arch_name: str,
        pkg_name: str,
        num_classes: int,
        resolution=224,
        arch_kwargs={},
    ) -> None:

        _output_transform = None

        def _output_transform(m: nn.Linear):
            def hook_fn(module, input, output):
                return BaseImageClassifierOutput(prediction=output, feature=input[0])

            m.register_forward_hook(hook_fn)

        try:
            tv_module = importlib.import_module(pkg_name)
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(f'{pkg_name} module not found.')
        factory = getattr(tv_module, arch_name, None)
        if factory is None:
            raise RuntimeError(f'{pkg_name} do not support model {arch_name}')
        model = factory(**arch_kwargs)

        feature_dim = reset_fc_layer(model, num_classes, _output_transform)

        super().__init__(resolution, num_classes, feature_dim)

        self.model = model

    def forward(self, image: torch.Tensor, *args, **kwargs):
        return self.model(image)


class BaseClassifierWrapperModel(BaseImageClassifier):

    def __init__(
        self, inner_model_name: str, inner_model_kwargs: Optional[dict] = None, **kwargs
    ):

        if inner_model_kwargs is None:
            inner_model_kwargs = {}

        factory = get_model_builder(inner_model_name)
        inner_model = factory(**inner_model_kwargs)

        super(BaseClassifierWrapperModel, self).__init__(
            inner_model.resolution, inner_model.num_classes, inner_model.feature_dim
        )

        self.inner_model = inner_model

    @staticmethod
    def create_from_model_impl(cls, inner_model: ModelMixin, **kwargs):
        inner_model_name = inner_model.__class__.__name__
        # inner_model_kwargs = inner_model.extract_init_dict(inner_model._internal_dict)
        inner_model_kwargs = extract_init_dict(inner_model)
        inner_model_kwargs = inner_model_kwargs[0]
        # print(inner_model_name, inner_model_kwargs.keys())
        wrapped_model = cls(
            inner_model_name=inner_model_name,
            inner_model_kwargs=inner_model_kwargs,
            **kwargs,
        )
        wrapped_model.inner_model.load_state_dict(inner_model.state_dict())

        return wrapped_model

    @classmethod
    def create_from_model(cls, inner_model: ModelMixin, **kwargs):
        return cls.create_from_model_impl(cls, inner_model, **kwargs)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from typing import Callable

import torch
from torch import nn
from torch.nn import functional as F

from ..auto import auto_register_save_load


def max_margin_loss(out, iden):
    real = out.gather(1, iden.unsqueeze(1)).squeeze(1)
    tmp1 = torch.argsort(out, dim=1)[:, -2:]
    new_y = torch.where(tmp1[:, -1] == iden, tmp1[:, -2], tmp1[:, -1])
    margin = out.gather(1, new_y.unsqueeze(1)).squeeze(1)

    return (-1 * real).mean() + margin.mean()


def poincare_loss(outputs, targets, xi=1e-4):
    # Normalize logits
    u = outputs / torch.norm(outputs, p=1, dim=-1).unsqueeze(1)
    # Create one-hot encoded target vector
    v = torch.clip(torch.eye(outputs.shape[-1])[targets.detach().cpu()] - xi, 0, 1)
    v = v.to(u.device)
    # Compute squared norms
    u_norm_squared = torch.norm(u, p=2, dim=1) ** 2
    v_norm_squared = torch.norm(v, p=2, dim=1) ** 2
    diff_norm_squared = torch.norm(u - v, p=2, dim=1) ** 2
    # Compute delta
    delta = 2 * diff_norm_squared / ((1 - u_norm_squared) * (1 - v_norm_squared))
    # Compute distance
    loss = torch.arccosh(1 + delta)
    return loss.mean()


_LOSS_MAPPING = {
    'ce': F.cross_entropy,
    'poincare': poincare_loss,
    'max_margin': max_margin_loss,
}


def register_loss_function(name: str, fn: Callable):
    _LOSS_MAPPING[name] = fn


@auto_register_save_load
class TorchLoss:
    """Find loss function from 'torch.nn.functional' and 'torch.nn'"""

    def __init__(self, loss_fn: str | Callable, **kwargs) -> None:
        # super().__init__()
        self.fn = None
        if isinstance(loss_fn, str):
            if loss_fn.lower() in _LOSS_MAPPING:
                self.fn = _LOSS_MAPPING[loss_fn.lower()]
            else:
                module = importlib.import_module('torch.nn.functional')
                fn = getattr(module, loss_fn, None)
                if fn is not None:
                    self.fn = lambda *arg, **kwd: fn(*arg, **kwd, **kwargs)
                else:
                    module = importlib.import_module('torch.nn')
                    t = getattr(module, loss_fn, None)
                    if t is not None:
                        self.fn = t(**kwargs)
                if self.fn is None:
                    raise RuntimeError(f'loss_fn {loss_fn} not found.')
        else:
            self.fn = loss_fn

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import datetime
import sys
import platform
import os
import subprocess
from dataclasses import dataclass
from typing import Dict

from .dataclass_utils import config_dataclass_wrapper

logger = logging.getLogger(__name__)


def get_current_commit_hash():
    try:
        result = subprocess.run(
            ["git", "rev-parse", "HEAD"], capture_output=True, text=True, check=True
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        return "N/A (due to not a git repository)"


def get_version_pip():
    result = subprocess.run(["pip", "list"], capture_output=True, text=True)
    packages_dict = {}
    lines = result.stdout.split('\n')
    for line in lines[2:]:
        if line:
            try:
                package_name, version, *_ = line.split()
                packages_dict[package_name] = version
            except:
                logger.warning(f"Failed to parse package version: {line}")
    return packages_dict


def clear_packages(origin_packages: Dict):

    result_packages = {}
    loaded_modules = sys.modules

    for module_name, module in dict(loaded_modules).items():
        if '.' in module_name:
            continue
        if module_name in origin_packages:
            result_packages[module_name] = origin_packages[module_name]
            if hasattr(module, '__version__'):
                result_packages[module_name] = module.__version__

    return result_packages


def get_version(clear_package: bool = False) -> dict:
    packages_dict = get_version_pip()
    if clear_package:
        packages_dict = clear_packages(packages_dict)
    return {
        "_datetime": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "_git_commit": get_current_commit_hash(),
        "_platform": platform.platform(),
        "_processor": platform.processor(),
        "_python": sys.version,
        "_python_compiler": platform.python_compiler(),
        "_python_packages": packages_dict,
    }


@config_dataclass_wrapper(config_name='version.json')
@dataclass
class VersionInfo:
    _datetime: str
    _git_commit: str
    _platform: str
    _processor: str
    _python: str
    _python_compiler: str
    _python_packages: dict

    @classmethod
    def create(cls, clear_package: bool = False):
        return cls(**get_version(clear_package=clear_package))



# import os
# from typing import Dict

# import torch
# from torch.utils.data import Dataset, DataLoader

# from dataclasses import dataclass


# @dataclass
# class DataManager:
#     dataset: Dataset
#     dataloader: DataLoader

#     def save_pretrained(self, save_directory: str) -> None:
#         """Save the dataset and dataloader to the specified directory."""
#         os.makedirs(save_directory, exist_ok=True)
#         torch.save(self.dataset, f"{save_directory}/dataset.pt")
#         torch.save(self.dataloader, f"{save_directory}/dataloader.pt")

#     @classmethod
#     def from_pretrained(cls, save_directory: str) -> "DataManager":
#         """Load the dataset and dataloader from the specified directory."""
#         dataset = torch.load(f"{save_directory}/dataset.pt")
#         dataloader = torch.load(f"{save_directory}/dataloader.pt")
#         return cls(dataset, dataloader)


# @dataclass
# class DataManagerGroup:
#     """Data Manager class."""

#     _datasets: Dict[str, DataManager] = None

#     def __delattr__(self, name: str) -> None:
#         del self._datasets[name]

#     def __setattr__(self, name: str, data_manager: DataManager) -> None:
#         self._datasets[name] = data_manager

#     def __getattr__(self, name: str) -> DataManager:
#         return self._datasets[name]

#     def __getitem__(self, name: str) -> DataManager:
#         return self._datasets[name]

#     def __setitem__(self, name: str, data_manager: DataManager) -> None:
#         self._datasets[name] = data_manager

#     def __delitem__(self, name: str) -> None:
#         del self._datasets[name]

#     def __contains__(self, name: str) -> bool:
#         return name in self._datasets

#     def save_pretrained(self, save_directory: str) -> None:
#         for name, data_manager in self._datasets.items():
#             data_manager.save_pretrained(f"{save_directory}/{name}")

#     @classmethod
#     def from_pretrained(cls, save_directory: str) -> "DataManagerGroup":
#         datasets = {}
#         for name in os.listdir(save_directory):
#             subfolder = os.path.join(save_directory, name)
#             if os.path.isdir(subfolder):
#                 datasets[name] = DataManager.from_pretrained(subfolder)
#         return cls(datasets)



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import heapq
import logging
from typing import AnyStr, Dict, Callable, Literal, TypeVar, Union, Generic
from dataclasses import dataclass

import pandas as pd

logger = logging.getLogger(__name__)

_highlight_format_mapping = {
    "none": "{}",
    "bold": "\\mathbf{{{}}}",
    "italic": "\\mathit{{{}}}",
    "underline": "\\underline{{{}}}",
}

_default_highlight = "bold"


def register_highlight_format(name: str, format: str):
    if name in _highlight_format_mapping:
        logger.warning(f"Overwriting existing highlight format: {name}")
    _highlight_format_mapping[name] = format
    logger.info(f"Registered highlight format: {name}: {format}")


def list_all_highlight_format():
    return list(_highlight_format_mapping.keys())


def _get_last_number(string: str):
    match = re.search(r'\d+$', string)
    if match:
        return int(match.group())
    return None


def get_highlight_type_function(highlight_type: str):
    if "max" in highlight_type:
        if highlight_type == "max":
            return max
        max_num = _get_last_number(highlight_type)
        if max_num is None:
            return max
        return lambda x: heapq.nlargest(max_num, x)[-1]
    if "min" in highlight_type:
        if highlight_type == "min":
            return min
        min_num = _get_last_number(highlight_type)
        if min_num is None:
            return min
        return lambda x: heapq.nsmallest(min_num, x)[-1]
    raise ValueError(f"Invalid highlight type string: {highlight_type}")


@dataclass
class DataPoint:
    mean: float
    std: float | None = None
    highlight_format: str = _default_highlight
    decimal: int = None
    mean_rounded: float | None = None
    std_rounded: float | None = None

    def __post_init__(self):
        # if self.decimal is not None:
        self.set_decimal(self.decimal)
        self.set_highlight_format(self.highlight_format)

    def _preprocess_highlight_format(self, highlight_format: str):
        if highlight_format is None:
            return _highlight_format_mapping["none"]
        highlight_format_lower = highlight_format.lower()
        if highlight_format_lower in _highlight_format_mapping:
            highlight_format = _highlight_format_mapping[highlight_format_lower]
            return highlight_format
        return highlight_format

    def set_highlight_format(self, highlight_format=None):
        self.highlight_format = self._preprocess_highlight_format(highlight_format)

    def set_decimal(self, decimal):
        if decimal is None:
            self.decimal = None
            self.mean_rounded = self.mean
            self.std_rounded = self.std
            return
        self.decimal = decimal
        if self.std is not None:
            self.std_rounded = round(self.std, decimal)
        self.mean_rounded = round(self.mean, decimal)

    # compare two distributions
    def __lt__(self, other: "DataPoint"):
        return self.mean_rounded < other.mean_rounded

    def __eq__(self, other: "DataPoint"):
        return self.mean_rounded == other.mean_rounded

    def __gt__(self, other: "DataPoint"):
        return self.mean_rounded > other.mean_rounded

    def __le__(self, other: "DataPoint"):
        return self.mean_rounded <= other.mean_rounded

    def __ge__(self, other: "DataPoint"):
        return self.mean_rounded >= other.mean_rounded

    def __str__(self):

        decimal = self.decimal

        mean_s = f"{self.mean:.{decimal}f}" if decimal is not None else str(self.mean)
        # if self.highlight:
        mean_s = self.highlight_format.format(mean_s)
        std_s = ""
        if self.std is not None:
            std_s = (
                f"\\pm {self.std:.{decimal}f}"
                if decimal is not None
                else f"\\pm {self.std}"
            )
            std_s = f"_{{{std_s}}}"
        return f"${mean_s}{std_s}$"


class ListDataPoint:

    def __init__(self, mean, std=None, decimal=None, highlight_type=None):
        self.points = (
            [DataPoint(m, s) for m, s in zip(mean, std)]
            if std is not None
            else [DataPoint(m) for m in mean]
        )
        self.decimal = decimal
        self.highlight_type = highlight_type
        for d in self.points:
            d.set_decimal(decimal)

        self.set_highlight(highlight_type)

    @classmethod
    def concat(cls, list_data_points: list["ListDataPoint"]):
        all_means = [
            p.mean
            for list_data_point in list_data_points
            for p in list_data_point.points
        ]
        all_stds = [
            p.std
            for list_data_point in list_data_points
            for p in list_data_point.points
        ]
        # check decimal and highlight_type is the same
        if not all(
            list_data_point.decimal == list_data_points[0].decimal
            for list_data_point in list_data_points
        ):
            raise ValueError("Decimal must be the same for all data points")

        if not all(
            list_data_point.highlight_type == list_data_points[0].highlight_type
            for list_data_point in list_data_points
        ):
            raise ValueError("Highlight type must be the same for all data points")
        return cls(
            all_means,
            all_stds,
            decimal=list_data_points[0].decimal,
            highlight_type=list_data_points[0].highlight_type,
        )

    def __len__(self):
        return len(self.points)

    def __iter__(self):
        return iter(self.points)

    def __getitem__(self, i):
        return self.points[i]

    def _preprocess_highlight_type(
        self,
        highlight_type: (
            None | str | Callable | Dict[AnyStr, AnyStr] | Dict[Callable, AnyStr]
        ),
    ):
        if highlight_type is None:
            return {}

        if not isinstance(highlight_type, dict):
            highlight_type = {highlight_type: "bold"}

        processed_highlight_type = {}
        for key, value in highlight_type.items():
            if isinstance(key, str):
                key = get_highlight_type_function(key)
            processed_highlight_type[key] = value
        return processed_highlight_type

    def set_highlight(self, highlight_type=None, overwrite: bool = True):
        highlight_type = self._preprocess_highlight_type(highlight_type)

        if overwrite:
            for d in self.points:
                d.set_highlight_format(_highlight_format_mapping["none"])

        for anchor_fn, highlight_format in highlight_type.items():
            anchor = anchor_fn(self.points)
            for d in self.points:
                if d == anchor:
                    d.set_highlight_format(highlight_format)

    def get_list_str(self):
        return [str(d) for d in self.points]

    @property
    def fixed(self):
        return self.get_list_str()


class TableDataPoint:

    def __init__(self, list_points: list[list[str], ListDataPoint]):
        self.list_points = list_points

        if not all(
            len(list_point) == len(self.list_points[0])
            for list_point in self.list_points
        ):
            raise ValueError(
                f"All lists in list_points must have the same length, but found {[len(list_point) for list_point in self.list_points]}"
            )

    @property
    def str_matrix(self):
        return [list(map(str, col)) for col in self.list_points]

    @property
    def transpose_str_matrix(self):
        return [list(map(str, col)) for col in zip(*self.list_points)]
        # return list(map(list, zip(*self.list_points)))

    def __str__(self):
        matrix = self.transpose_str_matrix
        return "".join([" & ".join(row) + " \\\\\n" for row in matrix])

    def __getitem__(self, indice):
        return self.list_points[indice]

    def get_col_at(self, indice):
        return self.__getitem__(indice)

    def get_row_at(self, indice):
        return [list_point[indice] for list_point in self.list_points]

    def __len__(self):
        return len(self.list_points)

    @property
    def num_cols(self):
        return len(self)

    @property
    def num_rows(self):
        return len(self.list_points[0])

    @staticmethod
    def concat(
        table_data_points: list["TableDataPoint"], axis: Literal['col', 'row'] = 'col'
    ):
        if axis == "col":
            return TableDataPoint(
                [
                    list_point
                    for table_data_point in table_data_points
                    for list_point in table_data_point.list_points
                ]
            )
        elif axis == "row":
            if not all(
                len(table_data_point) == len(table_data_points[0])
                for table_data_point in table_data_points
            ):
                raise ValueError("All TableDataPoint must have the same length")

            all_cols = []
            for i in range(len(table_data_points[0])):
                cols = [table_data_point[i] for table_data_point in table_data_points]
                if isinstance(cols[0], ListDataPoint):
                    all_cols.append(ListDataPoint.concat(cols))
                else:
                    all_cols.append([item for col in cols for item in col])
            return TableDataPoint(all_cols)
        else:
            raise ValueError("Axis must be 'col' or 'row'")

    @property
    def fixed(self):
        lists = []
        for list_point in self.list_points:
            if isinstance(list_point, ListDataPoint):
                lists.append(list_point.fixed)
            else:
                lists.append(list_point)
        return TableDataPoint(lists)


@dataclass
class MultiCellInfo:
    context: str
    length: int | None = None


# @dataclass
# class MultiRowInfo(MultiCellInfo):
#     pass


# @dataclass
# class MultiColInfo(MultiCellInfo):
#     pass


# T = TypeVar("T", bound=Union[MultiRowInfo, MultiColInfo])


class ListMultiCellInfo:
    def __init__(self, infos: list[MultiCellInfo]):
        all_infos = []
        for info in infos:
            if isinstance(info, str):
                info = MultiCellInfo(info, 1)
            elif isinstance(info, MultiCellInfo):
                pass
            else:
                info = MultiCellInfo(info[0], info[1])
            all_infos.append(info)

        self.infos = all_infos

        # print(len(self))

    # def __len__(self):
    #     return len(self.infos)

    def __getitem__(self, item):
        return self.infos[item]

    def __iter__(self):
        return iter(self.infos)

    @property
    def length(self):
        # return None
        current_length = 0
        for info in self.infos:
            if info.length is None:
                return None
            current_length += info.length
        return current_length

    def fill_length(self, length):
        none_indices = []
        current_length = 0
        for idx, info in enumerate(self.infos):
            if info.length is None:
                none_indices.append(idx)
            else:
                current_length += info.length
        if len(none_indices) == 0:
            if not current_length == length:
                raise ValueError(f"Invalid length: {length}, expected {current_length}")
            return self
        if len(none_indices) == 1:
            if current_length > length:
                raise ValueError(f"Invalid length: {length}, expected {current_length}")
            self.infos[none_indices[0]].length = length - current_length
            return
        # len(none_indices) > 1
        raise ValueError("None length is larger than 1. Fail to fill length")


class ListMultiRowInfo(ListMultiCellInfo):

    def create_table(self):
        contents = []
        for info in self.infos:
            if info.length is None:
                raise RuntimeError("None length is not allowed")
            if info.length == 1:
                contents.append(info.context)
            else:
                contents.append(f"\\multirow{{{info.length}}}{{*}}{{{info.context}}}")
                contents.extend([""] * (info.length - 1))
        return TableDataPoint([contents])


class TableRow:

    def __init__(self, contents: list[ListMultiRowInfo, TableDataPoint]):
        self.contents = contents
        self.length = self._check_length()

    def create_table(self):
        tables = [
            content.create_table() if isinstance(content, ListMultiRowInfo) else content
            for content in self.contents
        ]
        return TableDataPoint.concat(tables, axis="col")

    def _check_length(self):
        standard_length = None
        for content in self.contents:
            if isinstance(content, ListMultiCellInfo):
                current_length = content.length
            elif isinstance(content, TableDataPoint):
                current_length = content.num_rows
            else:
                raise ValueError("Invalid content")

            if (
                standard_length is not None
                and current_length is not None
                and standard_length != current_length
            ):
                raise ValueError(
                    f"Invalid length, current length is {current_length}, reference length is {standard_length}"
                )
            standard_length = (
                standard_length if standard_length is not None else current_length
            )

        if standard_length is None:
            raise ValueError("Invalid length")

        for content in self.contents:
            if isinstance(content, ListMultiCellInfo):
                content.fill_length(standard_length)

        return standard_length



import sys

import argparse

import argparse
import sys


from .cuda import _wait_gpu_run_register


def main():
    parser = argparse.ArgumentParser(
        description='PhoenixCat CLI',
        formatter_class=argparse.RawTextHelpFormatter,
    )

    subparsers = parser.add_subparsers(
        title='command',
        dest='command',  # 必须有，用于判断用户输入了哪个子命令
        required=True,
    )

    _wait_gpu_run_register(subparsers)

    args = parser.parse_args()

    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



from .dataloader import DataLoaderConfig, getDataLoader
from .dataset_wrappers import (
    DatasetWrapper,
    DownSampleDatasetWrapper,
    RandomDownSampleDatasetWrapper,
    DatasetIndexWrapper,
)
from .label_dataset import LabelImageFolder

# from .manager import DataManager, DataManagerGroup
